#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#ifndef USE_NETLIB
#include <net/ethernet.h>
#include <netinet/in.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>
#else
#include <netlib.h>
#endif

#include "pktlib.h"
#include "pktbuf.h"

struct macaddr_entry {
  struct macaddr_entry *next;
  struct ether_addr macaddr;
  pktif_t pktif;
};

static struct macaddr_entry *macaddr_table = NULL;

struct macaddr_entry *macaddr_entry_search(char macaddr[])
{
  struct macaddr_entry *entry;
  for (entry = macaddr_table; entry; entry = entry->next) {
    if (!memcmp(macaddr, &entry->macaddr, ETHER_ADDR_LEN))
      return entry;
  }
  return NULL;
}

struct macaddr_entry *macaddr_entry_create(char macaddr[])
{
  struct macaddr_entry *entry;

  entry = malloc(sizeof(*entry));
  memset(entry, 0, sizeof(*entry));

  memcpy(&entry->macaddr, macaddr, ETHER_ADDR_LEN);

  entry->next = macaddr_table;
  macaddr_table = entry;

  return entry;
}

struct macaddr_entry *macaddr_entry_add(char macaddr[], pktif_t pktif)
{
  struct macaddr_entry *entry;

  entry = macaddr_entry_search(macaddr);
  if (entry == NULL)
    entry = macaddr_entry_create(macaddr);

  entry->pktif = pktif;

  return entry;
}

int macaddr_entry_dump()
{
  struct macaddr_entry *entry;

  printf("---- MAC Address Table ----\n");
  for (entry = macaddr_table; entry; entry = entry->next) {
    printf("MAC Address: %s (Interface: %s)\n",
	   ether_ntoa(&entry->macaddr), pktif_get_name(entry->pktif));
  }
  fflush(stdout);

  return 0;
}

int main(int argc, char *argv[])
{
  pktif_t pktif, p;
  int r, rate = 0;
  pktbuf_t pktbuf;
  int type;
  struct header_param param;
  struct macaddr_entry *entry;

  argc--; argv++;

  if (isdigit(argv[0][0])) { /* ͤꤵ줿 */
    rate = atoi(argv[0]); /* ѥåѴΨ */
    argc--; argv++;
    srand(time(NULL));
  }

  while (argc > 0) { /* ʣ󥿡ե򥪡ץ󤹤 */
    /* 󥿡ե򥪡ץ󤹤(ȯѥåȤϼʤ) */
    pktif_open(argv[0], PKTIF_OPEN_FLAG_RECV_NOTSENT, 0);
    argc--; argv++;
  }

  while (1) {
    pktif = pktlib_iflist_select(1000000); /* ʣ󥿡եԤ */
    if (pktif == NULL) {
      macaddr_entry_dump();
      continue;
    }

    pktbuf = pktbuf_recv(pktif, 0); /* ѥåȤ */
    if (pktbuf == NULL)
      continue;

    if ((rate > 0) && ((rand() % rate) == 0))
      goto discard; /* ꤵ줿ΨǥѥåȤѴ */

    type = pktif_get_linktype(pktif);
    if (type != DLT_EN10MB)
      goto discard;

    r = pktbuf_analyze_ethernet(pktbuf, &param);
    if (r < 1)
      goto discard;

    /* MACɥ쥹ؽ */
    macaddr_entry_add(param.ethernet.src_macaddr, pktif);

    /* ѥåȤ */
    entry = macaddr_entry_search(param.ethernet.dst_macaddr);
    if ((param.ethernet.dst_macaddr[0] & 1) || (entry == NULL)) {
      /* ֥ɥ㥹ȤݡȤ */
      for (p = pktlib_iflist_get_list(); p; p = pktif_get_next(p)) {
	if (p != pktif)
	  pktbuf_send(p, pktbuf_clone(pktbuf));
      }
    } else {
      pktbuf = pktbuf_send(entry->pktif, pktbuf);
    }

discard:
    pktbuf_destroy(pktbuf);
  }

  return 0;
}
